{$IFNDEF FPC}
  {$N+}
  {$E+}
{$ENDIF}
{
        Program Name    : CVT
        Package Name    : Subroutine Library - Layer 1
        Purpose         : Data conversion routines
        Institution     : Conroy & Conroy Co.
        Date written    : 10-Dec-85
        Written by      : Alan Conroy
        Version         : 1.0E

        Copyright (C) 1985-2019 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

      ***************************************************************
      *                                                             *
      *      M O D I F I C A T I O N   H I S T O R Y                *
      *                                                             *
      ***************************************************************

          DATE        BY            REASON

        03-Sep-87     EAC           Disallow bases<2
        03-Dec-01     EAC           Added string/numeric
                                    conversion functions.
        13-Jun-02     EAC           Added Translate_Value.
        1-Nov-02      EAC           Support int64 parameters.
      				                      Added eicosahedron.
                                    Fix compilation errors in BP.
        23-Sep-05     EAC           Added Generic_String_To_Numeric
                                    and various other enhancements.
        21-Jul-2015   EAC           Added Hex.
        2-Nov-2018    EAC           Fix Hex (returning wrong values).
        19-Jun-2019   EAC           Handle 0-based strings.

      *******************************************************
      *                                                     *
      *          P R O G R A M   P U R P O S E              *
      *                                                     *
      *******************************************************

      This module contains routines for converting numeric data.

      CODE CERTIFIED - LEVEL 1 by EAC on 24-Sep-2005.  Is_Roman untested.  All untested on Mobile.
}

{TODO: move English-related conversions to DLL}

unit Cvt ;

interface

uses { Layer 0... }
     _Types ; { int64 }

{$I EDEFINES.INC}

{ Convert from any base to any base, provided that the source and destination
  bases are integer values.  Legal base values are zero (0) to 49.  Bases
  outside of this range may or may not yeild correct results. }
function Cvtb( Base1, Base2 : Integer ; X : Ansistring ) : Ansistring ;

{ Determines last position in S that is part of a number in English.  Returns
  < 1 if S doesn't start with a valid number in English. }
function English_Numeric_Extent( S : Ansistring ) : integer ;

function English_To_Numeric( S : Ansistring ; var Numeric : double ) : boolean ;

{ Converts a numeric string to a number and returns the type of the value.  The
  return is True if the passed numeric string is recognized or false
  otherwise. }
function English_String_To_Numeric( S : Ansistring ; var Numeric : double ;
    var Typ : integer ) : boolean ;

{ A fast hex converter.  Returns an integer whose lower 2 bytes contain the
  ASCII values which are the hex digits for the passed byte value. }
function FHex( Value : byte ) : integer ;

function From_Hex( Value : Ansistring ) : int64 ;

{ Converts a numeric string to a number and returns the type of the value.  The
  return is True if the passed numeric string is recognized or false
  otherwise.  This differs from String_To_Numeric in that it recognizes only
  strings containing valid prefix and suffixes, even if the specific combination
  is not a recognized English word.  For example, "centgram" would be recognized
  as a 100-sided 2D object.  However, strings without both prefix and suffix
  are not recognized (such as "one", or "hex"). }
function Generic_String_To_Numeric( S : Ansistring ; var Numeric : double ;
    var Typ : integer ) : boolean ;

function Hex( Value : byte ) : Ansistring ;

function Is_Roman( const S : string ) : boolean ;

{ Returns the minimum number of 8-bit bytes required to hold the specified value
  of the specified base. }
function Min_Bytes( const V : Ansistring ; Base : integer ) : integer ;

{ Converts a number to a specific numeric string. }
function Numeric_To_String( Numeric : double ; Typ : integer ) : Ansistring ;

{ Translates the passed string to an int64 and returns it as well as the base.
  Unless the passed string has a base modifier, the passed base is assumed to be
  the base of the passed value.  The function returns True if the value was
  valid, and false otherwise.  Flags are ignored for now. }
function Translate_Value( SValue : Ansistring ; Flags : integer ;
    var Base : integer ; var Value : int64 ) : boolean ;

{ Indicates whether passed string contains valid digits for the specified base
  from 0 to 49. }
function Valid_Base( Value : Ansistring ; Base : Integer ) : Boolean ;

{$IFDEF CC_Regression_Test}
procedure Test_Unit ;
{$ENDIF}


implementation

uses { Borland... }
{$IFDEF FPC_OR_DELPHI}
     SysUtils, { Exception }

{$ENDIF}
     { Layer 0... }
     L10n, { Text_* }
     CommonUt, { Low_String_Index }
     Hexs, { FHex }
     _UE, { PSUnified_Exception }
     UE, {  }

     { Layer 1... }
     Language ; { Get_Language }

{ Unit internal routines... }

{ API.. }

function From_Hex( Value : Ansistring ) : int64 ;

begin
    From_Hex := Hexs.From_Hex( Value ) ;
end ;


function Hex( Value : byte ) : Ansistring ;

begin
    Hex := Hexs.Hex( Value ) ;
end ;


{ A fast hex converter.  Returns an integer whose lower 2 bytes contain the
  ASCII values which are the hex digits for the passed byte value. }
function FHex( Value : byte ) : integer ;

begin
    FHex := Hexs.FHex( Value ) ;
end ;


function Is_Roman( const S : string ) : boolean ;

var I : integer ;

begin
    Result := True ; { Assume valid }
    for I := Low_String_Index to High_String_Index( S ) do
    begin
        if( pos( upcase( S[ I ] ), 'CILMVX' ) < Low_String_Index ) then
        begin
            Result := False ;
            exit ;
        end ;
    end ;
end ;


{ Returns the minimum number of 8-bit bytes required to hold the specified value
  of the specified base. }
function Min_Bytes( const V : Ansistring ; Base : integer ) : integer ;

var Temp : Ansistring ;

begin
    Temp := CVTB( Base, 16, V ) ;
    Min_Bytes := ( length( Temp ) + 1 ) div 2 ; { Number of bytes required for V }
end ;


{ Converts a number to a specific numeric string. }
function Numeric_To_String( Numeric : double ; Typ : integer ) : Ansistring ;

begin
    Numeric_To_String := Get_Language( '' )^.Numeric_To_String( Numeric, Typ ) ;
end ; { Numeric_To_String }


function English_Numeric_Extent( S : Ansistring ) : integer ;

begin
    English_Numeric_Extent := Get_Language( '' )^.Numeric_Extent( S ) ;
end ;


function English_To_Numeric( S : Ansistring ; var Numeric : double ) : boolean ;

begin
    English_To_Numeric := Get_Language( '' )^.To_Numeric( S, Numeric ) ;
end ;


function Generic_String_To_Numeric( S : Ansistring ; var Numeric : double ;
    var Typ : integer ) : boolean ;

begin
    Generic_String_To_Numeric :=
        Get_Language( '' )^.Generic_String_To_Numeric( S, Numeric, Typ ) ;
end ; { Generic_String_To_Numeric }


function English_String_To_Numeric( S : Ansistring ; var Numeric : double ;
    var Typ : integer ) : boolean ;

begin
    English_String_To_Numeric := Get_Language( '' )^.String_To_Numeric( S, Numeric, Typ ) ;
end ;


const Base_List = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ!"#$%&?()*+~|' ;

{ Translates the passed string to an int64 and returns it as well as the base.
  Unless the passed string has a base modifier, the passed base is assumed to be
  the base of the passed value.  The function returns True if the value was
  valid, and false otherwise.  Flags are ignored for now. }
function Translate_Value( SValue : Ansistring ; Flags : integer ;
    var Base : integer ; var Value : int64 ) : boolean ;

    procedure Strip( Y : Integer ) ; { Strip specifier and set Base }

    begin
        Base := Y ;
        SValue := Copy( SValue, Low_String_Index, Length( SValue ) - 1 ) ;
    end ;

var D : integer ;

begin
    D := length( SValue ) - 1 + Low_String_Index ;
    if( D < Low_String_Index ) then
    begin
        Translate_Value := False ;
        exit ;
    end ;
    if( SValue[ D ] = 'H' ) then
    begin
        Strip( 16 ) ;
    end else
    if( ( SValue[ D ] = 'D' ) and ( Base < 14 ) ) then
    begin
        Strip( 10 ) ;
    end else
    if( ( SValue[ D ] = 'B' ) and ( Base < 11 ) ) then
    begin
        Strip( 2 ) ;
    end else
    if( ( SValue[ D ] = 'O' ) or ( SValue[ D ] = 'Q' ) ) then
    begin
        Strip( 8 ) ;
    end else
    if( SValue[ D ] = '.' ) then
    begin
        Strip( 10 ) ;
    end  ;
    SValue := Edit( SValue, 32 ) ;
    if( not Valid_Base( SValue, Base ) ) then
    begin
        Translate_Value := False ;
        exit ;
    end ;
    Translate_Value := True ;
    SValue := CVTB( Base, 10, SValue ) ;
    val( SValue, Value, D ) ;
end ;


function Valid_Base( Value : Ansistring ; Base : Integer ) : Boolean ;
{ Return true if Value is a valid value for base Base. }

var H, I, I1 : Integer ;

begin
    H := length( Value ) - 1 + Low_String_Index ;
    if( ( H > 0 ) and ( ord( Value[ H ] ) = 0 ) ) then
    begin
        dec( H ) ; // Don't include trailing null
    end ;
    if( ( H > 36 ) or ( H < Low_String_Index ) ) then
    begin
        Valid_Base := false ;
        exit ;
    end ;
    for I := Low_String_Index to H do
    begin
        I1 := Pos( upcase( Value[ I ] ), Base_List ) ;
        if( ( I1 = 0 ) or ( I1 > Base ) ) then { Verify that the value is valid }
        begin
            Valid_Base := false ;
            exit ;
        end ;
    end ;
    Valid_Base := true ;
end ;


function CvtB( Base1, Base2 : Integer ; X : Ansistring ) : Ansistring ;

Var Err : Boolean ;
    E, I : Integer ;
    X1, Xx : Extended ;
    Power : Extended ;
    Y : Ansistring ;

begin
    X1 := 0 ;
    X := Edit( X, 32 ) ;
    Err := False ;
    if( ( Base1 < 2 ) or ( Base2 < 2 ) or ( Base1 > 49 ) or ( Base2 > 49 ) ) then
    begin
        Fatal_Error( Create_PSSimple_UE( -1, 10, 1, UE_Error, Text_Error_Invalid_Base, '' ) ) ;
    end ;
    while( copy( X, Low_String_Index, 1 ) = '0' ) do { Trim off leading zeroes }
    begin
        X := copy( X, 1 + Low_String_Index, length( X ) ) ;
    end ;
    if( Length( X ) = 0 ) then
    begin
        X := '0' ; { A null value is assumed to be zero }
    end else
    if( X[ 1 ] = '-' ) then { Negative value }
    begin
        CVTB := '-' + CVTB( Base1, Base2, copy( X, 1 + Low_String_Index, length( X ) ) ) ;
        exit ;
    end ;
    if not Valid_Base( X, Base1 ) then
    begin
        Fatal_Error( Create_PSSimple_UE( -1, 10, 1, UE_Error, Text_Error_Invalid_Base, '' ) ) ;
    end ;
    if( Base1 = Base2 ) then
    begin
        cvtB := X ;
        exit ;
    end ;
    if( Length( X ) > 36 ) then
    begin
        { FP math will only handle 36 digits of base 2 with accuracy, larger
          bases are even more restricted }
        X := Copy( X, Length( X ) - 36, 36 ) ;
    end ;
    if( Base1 = 10 ) then
    begin
        Val( X, X1, E )
    end else
    begin
        { if the base to convert from is ten, skip the conversion }
        Power := 1 ;
        for I := Length( X ) - 1 + Low_String_Index downto Low_String_Index do
        { Convert value to a floating point value }
        begin
            Xx := Power * ( Pos( Copy( X, I, 1 ), Base_List ) - 1 ) ;
            X1 := X1 + Xx ;
            Power := Power * Base1 ;
            if( Power > 9999999999999999999.0 ) then
            begin
                if( not Err ) then
                begin
                    Fatal_Error( Create_PSSimple_UE( -1, 10, 1, UE_Error, Text_Error_Numeric_Error, '' ) ) ;
                end ;
                Err := True ;
                break ;
            end ; { if( Power > 9999999999999999999.0 ) }
        end ; { for I := Length( X ) - 1 + Low_String_Index downto Low_String_Index }
    end ; { if( Base1 = 10 ) }
    if( ( X1 > 9999999999999999999.0 ) and ( not Err ) ) then
    begin
        { UNTESTED }
        Fatal_Error( Create_PSSimple_UE( -1, 10, 1, UE_Error, Text_Error_Numeric_Error, '' ) ) ;
    end ;
    if( Base2 = 10 ) then
    begin
        Str( X1 : 13 : 0, Y ) ;
        Y := Edit( Y, 2 ) ;
        { if the base to convert to is ten, skip the conversion }
    end else
    begin
        while( X1 > 9999999999999999999.0 ) do
        begin
            { UNTESTED }
            X1 := Int( X1 / Base2 ) ;
        end ;
        Y := '' ;
        I := 0 ;
        Power := 1 ;
        while( Power <= X1 ) do
        begin
            I := I + 1 ;
            Power := Power * Base2 ;
        end ;
        for I := I - 1 downto 0 do
        begin
            Power := Power / Base2 ;
            if( Power <> 0 ) then { Power = 0 UNTESTED }
            begin
                Xx := Int( X1 / Power ) ;
                if( Xx < Base2 ) then
                begin
                    if( ( Xx <> 0 ) or ( Length( Y ) <> 0 ) ) then { else UNTESTED }
                    begin
                        Y := Y + Copy( Base_List, Trunc( Xx + Low_String_Index ), 1 ) ;
                    end ;
                    if( Xx <> 0 ) then { else UNTESTED }
                    begin
                        X1 := X1 - Power * Xx ;
                    end ;
                end
            end
        end ;
        if( Length( Y ) = 0 ) then
        begin
            Y := '0' ;
        end ;
    end ;
    Cvtb := Y ; { Return result }
end ; { CvtB }



{$IFDEF CC_Regression_Test}
procedure Error ;

begin
    writeln( 'Error' ) ;
    halt ;
end ;


procedure Test_Cvtb ;

begin
    if( CvtB( 10, 16, '255' ) <> 'FF' ) then
    begin
        Error ;
    end ;
    if( CvtB( 10, 16, '' ) <> '0' ) then
    begin
        Error ;
    end ;
    if( CvtB( 10, 16, '1' ) <> '1' ) then
    begin
        Error ;
    end ;
    if( CvtB( 16, 10, 'FF' ) <> '255' ) then
    begin
        Error ;
    end ;
  {$IFDEF EXCEPTION}
    try
  {$ENDIF}
        CvtB( 0, 0, '' ) ;
  {$IFDEF EXCEPTION}
        Error ;
    except
    end ;
  {$ENDIF}
  {$IFDEF EXCEPTION}
    try
  {$ENDIF}
        CvtB( 2, 8, 'F' ) ;
  {$IFDEF EXCEPTION}
        Error ;
    except
    end ;
  {$ENDIF}
  {$IFDEF EXCEPTION}
    try
  {$ENDIF}
        CvtB( 16, 10, 'FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF' ) ;
  {$IFDEF EXCEPTION}
        Error ;
    except
    end ;
  {$ENDIF}
  {$IFDEF EXCEPTION}
    try
  {$ENDIF}
        CvtB( 10, 16, '99999999999999999999' ) ;
  {$IFDEF EXCEPTION}
        Error ;
    except
    end ;
  {$ENDIF}
end ;


procedure Test_FHex ;

var I : integer ;
    S : string[ 2 ] ;

begin
    S := '  ' ;
    I := FHex( 255 ) ;
    move( I, S[ 1 ], 2 ) ;
    if( S <> 'FF' ) then
    begin
        Error ;
    end ;
    I := FHex( 0 ) ;
    move( I, S[ 1 ], 2 ) ;
    if( S <> '00' ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Min_Bytes ;

begin
    if( Min_Bytes( '255', 10 ) <> 1 ) then
    begin
        Error ;
    end ;
    if( Min_Bytes( 'FFFF', 16 ) <> 2 ) then
    begin
        Error ;
    end ;
    if( Min_Bytes( 'FFFFFFFF', 16 ) <> 4 ) then
    begin
        Error ;
    end ;
    if( Min_Bytes( 'FFFFFFFFFFFF', 16 ) <> 6 ) then
    begin
        Error ;
    end ;
    if( Min_Bytes( 'FFFFFFFFFFFFFF', 16 ) <> 7 ) then
    begin
        Error ;
    end ;
    if( Min_Bytes( '1', 10 ) <> 1 ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Translate_Value ;

var B : integer ;
    V : int64 ;

begin
    B := 10 ;
    if( not Translate_Value( '10', 0, B, V ) ) then
    begin
        Error ;
    end ;
    if( V <> 10 ) then
    begin
        Error ;
    end ;
    if( Translate_Value( '1F', 0, B, V ) ) then
    begin
        Error ;
    end ;
    if( Translate_Value( '', 0, B, V ) ) then
    begin
        Error ;
    end ;
    if( not Translate_Value( '10H', 0, B, V ) ) then
    begin
        Error ;
    end ;
    if( V <> 16 ) then
    begin
        Error ;
    end ;
    if( B <> 16 ) then
    begin
        Error ;
    end ;
    B := 10 ;
    if( not Translate_Value( '10D', 0, B, V ) ) then
    begin
        Error ;
    end ;
    if( V <> 10 ) then
    begin
        Error ;
    end ;
    if( B <> 10 ) then
    begin
        Error ;
    end ;
    if( not Translate_Value( '10B', 0, B, V ) ) then
    begin
        Error ;
    end ;
    if( V <> 2 ) then
    begin
        Error ;
    end ;
    if( B <> 2 ) then
    begin
        Error ;
    end ;
    if( not Translate_Value( '11Q', 0, B, V ) ) then
    begin
        Error ;
    end ;
    if( V <> 9 ) then
    begin
        Error ;
    end ;
    if( B <> 8 ) then
    begin
        Error ;
    end ;
    if( not Translate_Value( '11.', 0, B, V ) ) then
    begin
        Error ;
    end ;
    if( V <> 11 ) then
    begin
        Error ;
    end ;
    if( B <> 10 ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Valid_Base ;

begin
    if( Valid_Base( '1F', 10 ) ) then
    begin
        Error ;
    end ;
    if( not Valid_Base( 'F', 16 ) ) then
    begin
        Error ;
    end ;
    if( Valid_Base( '', 10 ) ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Unit ;

begin
    Test_Translate_Value ;
    Test_FHex ;
    Test_Valid_Base ;
    Test_Cvtb ;
    Test_Min_Bytes ;
end ;
{$ENDIF}


end.
