{$IFNDEF FPC}
  {$N+}
{$ENDIF}
{
        Program Name : Cache
        Package Name : Subroutine Library - Layer 1
        Purpose      : Cache control classes
        Institution  : Conroy & Conroy Co.
        Date Written : 18-Jun-1994
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 1994-2008 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          This unit provides classes and functions for handling caches.
}

unit Cache ;

interface

{$I EDEFINES.INC}

uses { Borland... }
{$IFDEF BP}
     Objects, { TCollection }
{$ENDIF}
     { Layer 0... }
     Collect, { TCollection }
     _Types, { int64 }

     { Layer 1... }
     _Cache ; { TSCache }

type TSHeap_Cache = object( TSCache )
                        private { Instance data... }
                            Limit : longint ; { Largest amount of data we can cache, in bytes (0 = unlimited) }
                            Current : longint ; { Amount of data currently cached, in bytes }
                            Blocks : TCollection ; { Collection of cache blocks }

                        private { Internal utility routines... }
                            procedure Trim ;

                        public { API... }
                            function Read_Cache( var Data ;
                                Address, Size : longint ) : longint ; virtual ;
                            { Reads up to Size bytes from the cache for the specified
                              address.  Returns the number of bytes actually read (if
                              0 then the data was not cached).  Cached data is copied
                              to Data. }

                            procedure Write_Cache( const Data ;
                                Address, Size : longint ) ; virtual ;
                            { Writes to cache up to Size bytes from Data. }

                            procedure Clear ; virtual ; { Clears cache }

                            procedure Initialize ; virtual ; { Initialize object }

                            procedure Terminate ; virtual ; { Terminate object }

                            procedure Set_Max_Size( Size : longint ) ; virtual ;
                            { Sets maximum cache size in bytes.  If this is more
                              than is currently cached, some cache is thrown away. }

                            function Get_Max_Size : longint ; virtual ;
                            { Gets current maximum cache size in bytes. }

                            function Get_Size : longint ; virtual ;
                            { Returns amount of currently cached data, in bytes. }

                            function Get_Stats( Index : longint ;
                                var Address : longint ) : longint ;
                                virtual ;
                            { Returns the size of the data at cached index Index
                              and sets Address of the address of the cached
                              data. }

                            function Lock( Address : longint ) : boolean ; virtual ;
                            { Locks the data starting at the specified address into
                              memory so that it is not thrown away.  Returns false if
                              the specified data is not in the cache. }

                            function Unlock( Address : longint ) : boolean ; virtual ;
                            { Unlocks the data starting at the specified address in
                              memory.  Returns false if the specified data is not in
                              the cache, or if it is not locked. }
                    end ; { TSHeap_Cache }
     PSHeap_Cache = ^TSHeap_Cache ;


type TSHeap_Cache64 = object( TSCache64 )
                          private { Instance data... }
                              Limit : longint ; { Largest amount of data we can cache, in bytes (0 = unlimited) }
                              Current : longint ; { Amount of data currently cached, in bytes }
                              Blocks : TCollection ; { Collection of cache blocks }

                          private { Internal utility routines... }
                              procedure Trim ;

                          public { API... }
                              function Read_Cache( var Data ;
                                  Address, Size : int64 ) : longint ; virtual ;
                              { Reads up to Size bytes from the cache for the specified
                                address.  Returns the number of bytes actually read (if
                                0 then the data was not cached).  Cached data is copied
                                to Data. }

                              procedure Write_Cache( const Data ;
                                  Address, Size : int64 ) ; virtual ;
                              { Writes to cache up to Size bytes from Data. }

                              procedure Clear ; virtual ; { Clears cache }

                              procedure Initialize ; virtual ; { Initialize object }

                              procedure Terminate ; virtual ; { Terminate object }

                              procedure Set_Max_Size( Size : longint ) ;
                                  virtual ;
                              { Sets maximum cache size in bytes.  If this is more
                                than is currently cached, some cache is thrown away. }

                              function Get_Max_Size : longint ; virtual ;
                              { Gets current maximum cache size in bytes. }

                              function Get_Size : longint ; virtual ;
                              { Returns amount of currently cached data, in bytes. }

                              function Get_Stats( Index : longint ;
                                  var Address : int64 ) : longint ;
                                  virtual ;
                              { Returns the size of the data at cached index
                                index and sets Address of the address of the
                                cached data. }

                              function Lock( Address : int64 ) : boolean ;
                                  virtual ;
                              { Locks the data starting at the specified address into
                                memory so that it is not thrown away.  Returns false if
                                the specified data is not in the cache. }

                              function Unlock( Address : int64 ) : boolean ;
                                  virtual ;
                              { Unlocks the data starting at the specified address in
                                memory.  Returns false if the specified data is not in
                                the cache, or if it is not locked. }
                      end ; { TSHeap_Cache64 }
     PSHeap_Cache64 = ^TSHeap_Cache64 ;


implementation

type TCache_Block = record
                        Address : longint ; { Address being cached }
                        Data : pointer ; { Pointer to cached data }
                        Size : longint ; { Size of data at Data }
                        Locked : boolean ; { True if this block is locked into cache }
                    end ;
     PCache_Block = ^TCache_Block ;


{ THeap_Cache methods... }

{ Internal utility routines... }

procedure TSHeap_Cache.Trim ;

var Block : PCache_Block ;

begin
    if( Blocks.Count > 0 ) then
    begin
        Block := Blocks.At( 0 ) ;
        freemem( Block^.Data, Block^.Size ) ;
        Current := Current - Block^.Size ;
        freemem( Block, sizeof( Block^ ) ) ;
        Blocks.AtDelete( 0 ) ;
    end ;
end ;


{ API... }

{ NOTE: Currently this implementation requires that reads and writes always
  start on specific addresses (ie, no reads in the middle of previously written
  blocks, etc.) }
function TSHeap_Cache.Read_Cache( var Data ; Address, Size : longint ) : longint ;

var Block : PCache_Block ;
    Loop : integer ;

begin
    for Loop := 0 to Blocks.Count - 1 do
    begin
        Block := Blocks.At( Loop ) ;
        if( Block^.Address = Address ) then
        begin
            if( Size > Block^.Size ) then
            begin
                Size := Block^.Size ;
            end ;
            move( Block^.Data^, Data, Size ) ;
            Read_Cache := trunc( Size ) ;
            inc( Statistics.Hits ) ;
            exit ;
        end ;
    end ;

    Read_Cache := 0 ; { Failure }
    inc( Statistics.Misses ) ;
end ;


procedure TSHeap_Cache.Write_Cache( const Data ; Address, Size : longint ) ;

var Block : PCache_Block ;
    Loop : integer ;

begin
    if( ( Size > Limit ) and ( Limit > 0 ) ) then
    begin
        exit ; { This block is too big for our limit }
    end ;
    inc( Statistics.Writes ) ;
    if( Limit > 0 ) then
    begin
        { Make room }
        if( Current + Size > Limit ) then
        begin
            inc( Statistics.Turns ) ;
        end ;
        while( Current + Size > Limit ) do
        begin
            Trim ;
        end ;
    end ;

    { See if block already in cache... }
    for Loop := 0 to Blocks.Count - 1 do
    begin
        Block := Blocks.At( Loop ) ;
        if( Block^.Address = Address ) then { Found in cache }
        begin
            Blocks.AtDelete( Loop ) ; { Remove from list }
            Current := Current - Block^.Size ;
            if( Size <> Block^.Size ) then { Size mismatch }
            begin
                freemem( Block^.Data, Block^.Size ) ;
                freemem( Block, sizeof( Block^ ) ) ;
                break ;
            end else
            begin
                Blocks.Insert( Block ) ; { Put back into cache at end of list }
                Current := Current + Size ;
                move( Data, Block^.Data^, Size ) ;
                exit ;
            end ;
        end ;
    end ;

    { Add to cache... }
{$IFDEF EXCEPTIONS}
    try
{$ENDIF}
        new( Block ) ;
        fillchar( Block^, sizeof( Block^ ), 0 ) ;
{$IFDEF EXCEPTIONS}
    except
        exit ;
    end ;
{$ENDIF}
{$IFDEF EXCEPTIONS}
    try
{$ENDIF}
        getmem( Block^.Data, Size ) ;
{$IFDEF EXCEPTIONS}
    except
        Block^.Data := nil ;
    end ;
{$ENDIF}
    if( Block^.Data = nil ) then { Out of memory }
    begin
        freemem( Block, sizeof( Block^ ) ) ;
        exit ;
    end ;
    Block^.Address := Address ;
    Block^.Size := Size ;
    move( Data, Block^.Data^, Size ) ;
    Blocks.Insert( Block ) ;
    Current := Current + Size ;
end ; { TSHeap_Cache.Write_Cache }


procedure TSHeap_Cache.Clear ;

begin
    while( Current > 0 ) do
    begin
        Trim ;
    end ;
end ;


procedure TSHeap_Cache.Initialize ;

begin
    Limit := 0 ;
    Current := 0 ;
    Blocks.Init( 0, 4 ) ;
    fillchar( Statistics, sizeof( Statistics ), 0 ) ;
end ;


procedure TSHeap_Cache.Terminate ;

begin
    Clear ;
    Blocks.Done ;
end ;


procedure TSHeap_Cache.Set_Max_Size( Size : longint ) ;

begin
    Limit := Size ;
    if( Limit > 0 ) then
    begin
        while( Limit < Current ) do { Trim data to fit new limit }
        begin
            Trim ;
        end ;
    end ;
end ;


function TSHeap_Cache.Get_Max_Size : longint ;

begin
    Get_Max_Size := Limit ;
end ;


function TSHeap_Cache.Get_Size : longint ;

begin
    Get_Size := Current ;
end ;


function TSHeap_Cache.Get_Stats( Index : longint ;
    var Address : longint ) : longint ;

var Block : PCache_Block ;

begin
    Get_Stats := 0 ;
    Address := -1 ;
    if( Blocks.Count > Index ) then
    begin
        Block := Blocks.At( Index ) ;
        Get_Stats := Block^.Size ;
        Address := Block^.Address ;
    end ;
end ;


function TSHeap_Cache.Lock( Address : longint ) : boolean ;

begin
    Lock := False ; { Not currently supported }
end ;


function TSHeap_Cache.Unlock( Address : longint ) : boolean ;

begin
    Unlock := False ; { Not currently supported }
end ;



type TCache64_Block = record
                          Address : int64 ; { Address being cached }
                          Data : pointer ; { Pointer to cached data }
                          Size : longint ; { Size of data at Data }
                          Locked : boolean ; { True if this block is locked into cache }
                      end ;
     PCache64_Block = ^TCache_Block ;


{ TSHeap_Cache64 methods... }

{ Internal utility routines... }

procedure TSHeap_Cache64.Trim ;

var Block : PCache64_Block ;

begin
    if( Blocks.Count > 0 ) then
    begin
        Block := Blocks.At( 0 ) ;
        freemem( Block^.Data, Block^.Size ) ;
        Current := Current - Block^.Size ;
        freemem( Block, sizeof( Block^ ) ) ;
        Blocks.AtDelete( 0 ) ;
    end ;
end ;


{ API... }

{ NOTE: Currently this implementation requires that reads and writes always
  start on specific addresses (ie, no reads in the middle of previously written
  blocks, etc.) }
function TSHeap_Cache64.Read_Cache( var Data ; Address, Size : int64 ) : longint ;

var Block : PCache64_Block ;
    Loop : integer ;
    S : longint ;

begin
{$IFDEF INT64}
    S := Size and $7FFF ;
{$ELSE}
    S := trunc( Size ) ;
{$ENDIF}
    for Loop := 0 to Blocks.Count - 1 do
    begin
        Block := Blocks.At( Loop ) ;
        if( Block^.Address = Address ) then
        begin
            if( S > Block^.Size ) then
            begin
                S := Block^.Size ;
            end ;
            move( Block^.Data^, Data, S ) ;
            Read_Cache := S ;
            inc( Statistics.Hits ) ;
            exit ;
        end ;
    end ;

    Read_Cache := 0 ; { Failure }
    inc( Statistics.Misses ) ;
end ;


procedure TSHeap_Cache64.Write_Cache( const Data ; Address, Size : int64 ) ;

var Block : PCache64_Block ;
    Loop : integer ;
    S : longint ;

begin
    if( ( Size > Limit ) and ( Limit > 0 ) ) then
    begin
        exit ; { This block is too big for our limit }
    end ;
    inc( Statistics.Writes ) ;
    if( Limit > 0 ) then
    begin
        { Make room }
        if( Current + Size > Limit ) then
        begin
            inc( Statistics.Turns ) ;
        end ;
        while( Current + Size > Limit ) do
        begin
            Trim ;
        end ;
    end ;
{$IFDEF INT64}
    S := Size and $7FFF ;
{$ELSE}
    S := trunc( Size ) ;
{$ENDIF}

    { See if block already in cache... }
    for Loop := 0 to Blocks.Count - 1 do
    begin
        Block := Blocks.At( Loop ) ;
        if( Block^.Address = Address ) then { Found in cache }
        begin
            Blocks.AtDelete( Loop ) ; { Remove from list }
            Current := Current - Block^.Size ;
            if( Size <> Block^.Size ) then { Size mismatch }
            begin
                freemem( Block^.Data, Block^.Size ) ;
                freemem( Block, sizeof( Block^ ) ) ;
                break ;
            end else
            begin
                Blocks.Insert( Block ) ; { Put back into cache at end of list }
                Current := Current + S ;
                move( Data, Block^.Data^, S ) ;
                exit ;
            end ;
        end ;
    end ; { for Loop := 0 to Blocks.Count - 1 }

    { Add to cache... }
{$IFDEF EXCEPTIONS}
    try
{$ENDIF}
        new( Block ) ;
        fillchar( Block^, sizeof( Block^ ), 0 ) ;
{$IFDEF EXCEPTIONS}
    except
        exit ;
    end ;
{$ENDIF}
{$IFDEF EXCEPTIONS}
    try
{$ENDIF}
        getmem( Block^.Data, S ) ;
{$IFDEF EXCEPTIONS}
    except
        Block^.Data := nil ;
    end ;
{$ENDIF}
    if( Block^.Data = nil ) then { Out of memory }
    begin
        freemem( Block, sizeof( Block^ ) ) ;
        exit ;
    end ;
{$IFDEF INT64}
    Block^.Address := Address and $7FFF ;
{$ELSE}
    Block^.Address := trunc( Address ) ;
{$ENDIF}
    Block^.Size := S ;
    move( Data, Block^.Data^, S ) ;
    Blocks.Insert( Block ) ;
    Current := Current + S ;
end ; { TSHeap_Cache64.Write_Cache }


procedure TSHeap_Cache64.Clear ;

begin
    while( Current > 0 ) do
    begin
        Trim ;
    end ;
end ;


procedure TSHeap_Cache64.Initialize ;

begin
    Limit := 0 ;
    Current := 0 ;
    Blocks.Init( 0, 4 ) ;
    fillchar( Statistics, sizeof( Statistics ), 0 ) ;
end ;


procedure TSHeap_Cache64.Terminate ;

begin
    Clear ;
    Blocks.Done ;
end ;


procedure TSHeap_Cache64.Set_Max_Size( Size : longint ) ;

begin
    Limit := Size ;
    if( Limit > 0 ) then
    begin
        while( Limit < Current ) do { Trim data to fit new limit }
        begin
            Trim ;
        end ;
    end ;
end ;


function TSHeap_Cache64.Get_Max_Size : longint ;

begin
    Get_Max_Size := Limit ;
end ;


function TSHeap_Cache64.Get_Size : longint ;

begin
    Get_Size := Current ;
end ;


function TSHeap_Cache64.Get_Stats( Index : longint ;
    var Address : int64 ) : longint ;

var Block : PCache64_Block ;

begin
    Get_Stats := 0 ;
    Address := -1 ;
    if( Blocks.Count > Index ) then
    begin
        Block := Blocks.At( Index ) ;
        Get_Stats := Block^.Size ;
        Address := Block^.Address ;
    end ;
end ;


function TSHeap_Cache64.Lock( Address : int64 ) : boolean ;

begin
    Lock := False ; { Not currently supported }
end ;


function TSHeap_Cache64.Unlock( Address : int64 ) : boolean ;

begin
    Unlock := False ; { Not currently supported }
end ;



end.
