{
        Unit Name    : HeapBuff
        Package Name : Subroutine Library - Layer 1
        Purpose      : Heap Buffer
        Institution  : Conroy & Conroy Co.
        Date Written : 29-Sep-2015
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (c) 2015 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          THeap_Buffer is a contiguous Buffer whose data resides on a heap.
}

unit HeapBuff ;

interface

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

uses // Subroutine Library - Layer 0...
     _Buffers, // TBuffer
     Heaps, // THeap
     _Streams, // TStream
     _UE, // TUnified_Exception

     // Subroutine Library - Layer 1...
     _Stores ; // TStore_Address

const THeap_Buffer_Facility_ID = 132 ;

type THeap_Buffer = class( TBuffer )
                        public // Constructors and destructors...
                            destructor Destroy ; override ;

                        private // Instance...
                            _Data : longint ;
                            _Heap : THeap ;
                            _Size : longint ;

                        protected // Property handlers...
                            function Get_Heap : THeap ; virtual ;
                            procedure Set_Heap( H : THeap ) ; virtual ;
                            function Get_Size : longint ; override ;
                            procedure Set_Size( _S : longint ) ; override ;

                        public // API...
                            function Add( Buf : PChar ; Len : longint ) : longint ;
                                override ;
                            procedure Read( PTR, Size : TStore_Address ;
                                var Buffer ) ; override ;
                            function Read_Data( var Data ; Address, _Size : TStore_Address ;
                                var UEC : TUnified_Exception ) : TStore_Address ;
                                override ;
                            function Load( Stream : TStream ) : boolean ;
                                override ;
                            function Persist( Stream : TStream ) : boolean ;
                                override ;
                            procedure Write( PTR, Size : TStore_Address ;
                                var Buffer ) ; override ;
                            function Write_Data( var Data ; Address,
                                _Size : TStore_Address ;
                                var UEC : TUnified_Exception ) : TStore_Address ;
                                override ;

                        published // Properties...
                            property Heap : THeap
                                read Get_Heap
                                write Set_Heap ;
                    end ; // THeap_Buffer

implementation

// THeap_Buffer methods...

// Constructors and destructors...

destructor THeap_Buffer.Destroy ;

begin
    Size := 0 ;
    Heap := nil ;

    inherited Destroy ;
end ;


// Property handlers...

function THeap_Buffer.Get_Heap : THeap ;

begin
    if( _Heap = nil ) then
    begin
        Heap := THeap.Create ;
    end ;
    Result := _Heap ;
end ;


procedure THeap_Buffer.Set_Heap( H : THeap ) ;

begin
    if( H <> nil ) then
    begin
        H.Attach ;
    end ;
    if( _Heap <> nil ) then
    begin
        _Heap.Detach ;
    end ;
    _Heap := H ;
end ;


function THeap_Buffer.Get_Size : longint ;

begin
    Result := _Size ;
end ;


procedure THeap_Buffer.Set_Size( _S : longint ) ;

var P : longint ;

begin
    if( _S <> _Size ) then
    begin
        P := Heap.Reallocate( _Data, _Size, _S ) ;
        if( P = 0 ) then
        begin
            exit ;
        end ;
        _Data := P ;
    end ;
end ;


// API...

function THeap_Buffer.Add( Buf : PChar ; Len : longint ) : longint ;

var Old : longint ;

begin
    Result := -1 ; // Assume failure to allocate more buffer
    if( Len < 1 ) then
    begin
        exit ;
    end ;
    Old := Size ;
    Size := Old + Len ;
    if( Size = Old + Len ) then // No change
    begin
        exit ;
    end ;
    Heap.Write( _Data + Old, Len, Buf ) ;
    Result := Old ;
end ;


function THeap_Buffer.Load( Stream : TStream ) : boolean ;

var B : byte ;
    Buff : array[ 0..32767 ] of byte ;
    Base, Count, Dummy, L : integer ;

begin
    Result := False ; // Assume failure

    // Read header (1 byte ID, 1 byte version, 4 byte lengtgh)...
    if( Stream = nil ) then
    begin
        exit ;
    end ;
    Dummy := Stream.Read( B, 1 ) ;
    if( Dummy <> 1 ) then
    begin
        exit ;
    end ;
    if( B <> THeap_Buffer_Facility_ID ) then // Not a persisted buffer
    begin
        exit ;
    end ;
    Dummy := Stream.Read( B, 1 ) ;
    if( Dummy <> 1 ) then
    begin
        exit ;
    end ;
    if( B <> 0 ) then // Not version 0
    begin
        exit ;
    end ;
    Dummy := Stream.Read( L, 4 ) ; // Get buffer length
    if( Dummy <> 4 ) then
    begin
        exit ;
    end ;

    // Read buffer contents from stream...
    Size := L ; // Create buffer
    Base := 0 ;
    while( L > 0 ) do
    begin
        Count := L ;
        if( Count > sizeof( Buff ) ) then
        begin
            Count := sizeof( Buff ) ;
        end ;
        Dummy := Stream.Read( Buff, Count ) ;
        if( Count <> Dummy ) then
        begin
            exit ;
        end ;
        L := L - Count ;
        Heap.Write( Base, Count, Buff) ;
        Base := Base + Count ;
    end ; // while( L > 0 )

    Result := True ;
end ; // THeap_Buffer.Load


function THeap_Buffer.Persist( Stream : TStream ) : boolean ;

var B : byte ;
    Buff : array[ 0..32767 ] of byte ;
    Base, Count, Dummy, Len : integer ;

begin
    Result := False ; // Assume failure
    if( Stream = nil ) then
    begin
        exit ;
    end ;

    // Read header (1 byte ID, 1 byte version, 4 byte lengtgh)...
    if( Stream = nil ) then
    begin
        exit ;
    end ;
    B := THeap_Buffer_Facility_ID ;
    Dummy := 1 ;
    Stream.Write( B, Dummy ) ; // Write facility ID
    if( Dummy <> 1 ) then
    begin
        exit ;
    end ;
    B := 0 ;
    Dummy := 1 ;
    Stream.Write( B, Dummy ) ; // Version 0
    if( Dummy <> 1 ) then
    begin
        exit ;
    end ;
    Len := Size ;
    Dummy := Stream.Write( Len, 4 ) ; // Buffer length

    // Write buffer contents to stream...
    Base := 0 ;
    while( Len > 0 ) do
    begin
        Count := Len ;
        if( Count > sizeof( Buff ) ) then
        begin
            Count := sizeof( Buff ) ;
        end ;
        if( Heap.Read( Base, Count, Buff ) <> Count ) then
        begin
            exit ;
        end ;
        Dummy := Count ;
        Stream.Write( Buff, Dummy ) ;
        if( Dummy <> Count ) then
        begin
            exit ;
        end ;
        Len := Len - Count ;
        Base := Base + Count ;
    end ; // while( Len > 0 )
    Result := True ;
end ; // THeap_Buffer.Persist


procedure THeap_Buffer.Read( PTR, Size : TStore_Address ; var Buffer ) ;

begin
    if( ( PTR < 0 ) or ( PTR >= _Size ) ) then
    begin
        exit ;
    end ;
    if( PTR + Size > _Size ) then
    begin
        Size := _Size - PTR - 1 ;
    end ;
    Heap.Read( PTR, Size, Buffer ) ;
end ;


function THeap_Buffer.Read_Data( var Data ; Address, _Size : TStore_Address ;
    var UEC : TUnified_Exception ) : TStore_Address ;

begin
    Read( Address, Size, Data ) ;
    UEC := nil ;
    Result := Size ;
end ;


procedure THeap_Buffer.Write( PTR, Size : TStore_Address ; var Buffer ) ;

begin
    if( ( PTR < 0 ) or ( PTR >= _Size ) ) then
    begin
        exit ;
    end ;
    if( PTR + Size > _Size ) then
    begin
        Size := _Size - PTR - 1 ;
    end ;
    Heap.Write( PTR, Size, Buffer ) ;
end ;


function THeap_Buffer.Write_Data( var Data ; Address, _Size : TStore_Address ;
    var UEC : TUnified_Exception ) : TStore_Address ;

begin
    Write( Address, _Size, Data ) ;
    UEC := nil ;
    Result := _Size ;
end ;


end.
