UNIT LSCCPU0;

{ NAME        : LSCCPU0
  DESCRIPTION : Determine on what type of CPU the system is running.

  The WhatCPU function examines the CPU and returns a number corresponding to
  the CPU type;  1 for 8086, 3 for 80386, etc.  This procedure came right
  out of Neil Rubenking's Turbo Pascal 6.0 Techniques and Utilities.
}

INTERFACE

CONST
  Cpu8086        = 1;
  Cpu80286       = 2;
  Cpu80386       = 3;
  Cpu80486       = 4;

  CPUIdToWORD    : ARRAY[Cpu8086..Cpu80486] OF WORD = (086, 286, 386, 486);


FUNCTION WhatCPU : WORD;


IMPLEMENTATION

  FUNCTION WhatCPU; ASSEMBLER;
  { WhatCPU determines the system processor by executing some common
    assembly language instructions, and testing the results.  The key is
    that some of the instructions behave differently on different processors.

    For example, 286, 386, and 486 class processors have different numbers
    of processor flags.  Thus, pushing and popping flags can be used to
    determine a processor by whether or not the flags are lost by masking the
    high order words with known values during the process.

    This function initially assumes an 8086 compatible is in use, and jumps
    to the exit point as soon as a processor type is known.

    Data Byte (DB) declarations are used in order to declare instructions
    which cannot be compiled in 8086 compatible mode, as our baseline is
    currently compiled.  Adding a "DB 66h" changes a 286 instruction into a
    386 extended instruction.  One instruction, the extended absolute AND, is
    coded completely in Data Bytes, as it cannot be compiled in less than
    386 mode.

    Note that this function currently only tests up to 486 class machines.
    For current usage, it is not important if the machine is more powerful
    than a 486.  An additional test, probably using the flag masking
    technique, will be added when we need to know that we are using a 586,
    or 686, etc.  }

  ASM
            MOV     DX,Cpu8086  { Initialize to XT type                     }

            PUSH    SP          { XT's push SP after it is decremented, but }
            POP     AX          { 286 and later push the current value, so  }
            CMP     SP,AX       { compare and see if it is an XT.           }
            JNE     @OUT

            MOV     DX,Cpu80286 { Initialize to 286 type                    }

            PUSHF               { 386 PUSHF and POPF instructions use a     }
            POP     AX          { 32 bit word size in order to include the  }
            OR      AX,4000h    { extended flags.  Thus, adding a mask at   }
            PUSH    AX          { the 26th bit (4000h) will test if this is }
            POPF                { a 286 when the flags are popped back off  }
            PUSHF               { and compared and the 26th bit is lost.    }
            POP     AX
            TEST    AX,4000h
            JE      @OUT

            MOV     DX,Cpu80386 { Initialize to 386 type                    }

                                { As with the 286 verus 386 above, use the  }
                                { 486 extended flags to test for a 386 POPFD}
                                { masking the 486 flags.                    }

    DB 66h; MOV     BX,SP       { MOV EBX,ESP                               }
    DB 66h, 83h, 0E4h, 0FCh     { AND ESP,FFFC                              }
    DB 66h; PUSHF               { PUSHFD                                    }
    DB 66h; POP     AX          { POP EAX                                   }
    DB 66h; MOV CX, AX          { MOV ECX,EAX                               }
    DB 66h, 35h, 00h
    DB 00h, 04h, 00             { XOR EAX,00040000h                         }
    DB 66h; PUSH    AX          { PUSH EAX                                  }
    DB 66h; POPF                { POPFD                                     }
    DB 66h; PUSHF               { PUSHFD                                    }
    DB 66h; POP     AX          { POP EAX                                   }
    DB 66h, 25h,00h
    DB 00h, 04h,00h             { AND EAX,00040000h                         }
    DB 66h, 81h,0E1h,00h
    DB 00h, 04h,00h             { AND ECX,00040000                          }
    DB 66h; CMP     AX,CX       { CMP EAX,ECX                               }
            JE      @Not486
            MOV     DX,Cpu80486
  @Not486:
    DB 66h; PUSH    CX          { PUSH ECX                                  }
    DB 66h; POPF                { POPFD                                     }
    DB 66h; MOV     SP,BX       { MOV ESP,EBX                               }
  @OUT:
            MOV     AX,DX
  END;

END.
