/*
        Program Name : Stores
        Package Name : Subroutine Library - Layer 1
        Purpose      : Base Storage objects
        Institution  : Conroy & Conroy Co.
        Date Written : 1999
        Written By   : Alan Conroy
        Version      : 1.1B

        Copyright (C) 1999-2016 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

          1-Nov-2002    EAC         Use TObject from Common.
         13-Nov-2002    EAC         Rename Tallocator to
                                    TManaged_Store.
         16-Nov-2005    EAC         Changed const parameters to var
                                    to compile with BP.

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TBase_Store is the base class for all storage-related classes.  It
        implements the generic store functionality upon which descendant stores
        are built.
*/

#ifndef STORESH
	#define STORESH

	#include "eDefines.h"
	#include "_cache.h"
	#include "_stores.h"
	#include "_ue.h"

const int StoreErr_Facility = 48 ;

const int StoreErr_Success = 0 ;
const int StoreErr_Invalid_Operation = 1 ; // The operation is not valid
const int StoreErr_Abstract = 2 ;

struct TStore_Statistics
{
    int Bytes_Read ; // Data bytes read
    int Bytes_Written ; // Data bytes written
    int Reads ; // Data read count
    int Writes ; // Data write count
    int Error_Count ; // Error count
} ;

class TStandard_COM_Store : public TCOM_Store
{
                               private: // Instance data...
                                   TStore_Statistics Store_Statistics ;
                                   TCOM_Cache* _Cache ;
                                   bool _Read_Only ;
                                   bool _Write_Only ;

                               public: // API...
                                   bool Is_Class( char* Name ) ;
                                       
                                   TStore_Address Read_Data( void &Data ;
                                       TStore_Address Address, TStore_Address _Size,
                                       TUnified_Exception* &UEC ) ;
                                   /* Read data from specified Address in store to Data.
                                     Read _Size bytes.  Returns success of operation in UEC.
                                     Function returns the number of bytes actually read. */

                                   TStore_Address Write_Data( void &Data ;
                                       TStore_Address Address, TStore_Address _Size,
                                       TUnified_Exception* &UEC ) ;
                                   /* Write data from Data to specified Address in store.
                                     Write _Size bytes.  Returns success of operation in UEC.
                                     Function returns the number of bytes actually written. */

                                   TStore_Size Max_Storage() ;
                                   /* Returns the maximum (current) size of the store, in
                                     bytes.  In other words, this is the highest valid
                                     address (plus 1). */

                                   TStore_Address Min_Storage() ;
                                   /* Returns the minimum atomic size of the store.  For
                                     instance, this would be the sector size for a disk. */

                                   TStore_Address Extend( TStore_Address Amount ) ;
                                   /* Requests extension of store by the amount specified.
                                     Returns the amount actually added to store.  If Amount
                                     is 0, function returns 0 if store is extendable, and -1
                                     if it is not. */

                                   bool Get_Read_Only() ;
                                   // Indicates that the store can only be read from.

                                   bool Get_Write_Only() ;
                                   // Indicates that the store can only be written to.

                                   void Format() ;
                                   // Does a low-level format of the store.

                                   char* Get_Name() ;
                                   // Returns name of store.

                                   TCOM_Cache* Get_Cache() ;
                                   // Returns current cache object.

                                   void Set_Cache( TCOM_Cache* Value ) ;
                                   // Sets the current cache object.

                                   bool Contiguous_Store() ;
                                   /* Returns True if the store's addresses are contiguous
                                     (non-sparse). */

                                   void Set_Max_Storage( TStore_Address Value,
                                       TUnified_Exception* &Res ) ;
                                   /* Sets the maximum (current) size of the store, in bytes.
                                     In other words, this is the new highest valid address
                                     (plus 1).  If the store is not resizable, the call
                                     fails and an error code is returned in Res.  This should
                                     only be used to shrink the size of the store, not to
                                     extend it.  To extend, use the Extend method. */

                                   TStore_Address Extended_Size() ;
                                   /* Returns the maximum theoretical size that the store can
                                     be extended to.  If the store is not extendable, this
                                     returns the same value as Max_Storage. */

                                   long int Get_Bytes_Read() ;
                                       // Data bytes read
                                   long int Get_Bytes_Written() ;
                                       // Data bytes written
                                   long int Get_Reads() ;
                                       // Data read count
                                   long int Get_Writes() ;
                                       // Data write count
                                   long int Get_Error_Count() ;
                                       // Error count
                                   void Set_Bytes_Read( long int Value ) ;
                                       // Data bytes read
                                   void Set_Bytes_Written( long int Value ) ;
                                       // Data bytes written
                                   void Set_Reads( long int Value ) ;
                                       // Data read count
                                   void Set_Writes( long int Value ) ;
                                       // Data write count
                                   void Set_Error_Count( long int Value ) ;
                                       // Error count
                                   void Set_Read_Only( bool Value ) ;
                                   // Sets the store read-only value.
                                   void Set_Write_Only( bool Value ) ;
                                   // Sets the store write-only value.
}

class TStandard_COM_Store64 : public TCOM_Store64
{
                                 private: // Instance data...
									 TStore_Statistics Store_Statistics ;
                                 	 TCOM_Cache64* _Cache ;
                                     bool _Read_Only  ;
                                     bool _Write_Only ;

                                 public: // API...
                                     bool Is_Class( char* Name ) ;
                                         
                                     TStore_Address64 Read_Data( void &Data, TStore_Address64 Address, TStore_Address64 _Size,
                                         TUnified_Exception* &UEC ) ;
                                     /* Read data from specified Address in store to Data.
                                       Read _Size bytes.  Returns success of operation in UEC.
                                       Function returns the number of bytes actually read. */

                                     TStore_Address64 Write_Data( void &Data, TStore_Address64 Address, TStore_Address64 _Size,
                                         TUnified_Exception* &UEC ) ;
                                     /* Write data from Data to specified Address in store.
                                       Write _Size bytes.  Returns success of operation in UEC.
                                       Function returns the number of bytes actually written. */

                                     TStore_Size64 Max_Storage() ;
                                     /* Returns the maximum (current) size of the store, in
                                       bytes.  In other words, this is the highest valid
                                       address (plus 1). */

                                     TStore_Address64 Min_Storage() ;
                                     /* Returns the minimum atomic size of the store.  For
                                       instance, this would be the sector size for a disk. */

                                     TStore_Address64 Extend( Amount : TStore_Address64 ) ;
                                     /* Requests extension of store by the amount specified.
                                       Returns the amount actually added to store.  If Amount
                                       is 0, function returns 0 if store is extendable, and -1
                                       if it is not. */

                                     bool Get_Read_Only() ;
                                     // Indicates that the store can only be read from.

                                     bool Get_Write_Only() ;
                                     // Indicates that the store can only be written to.

                                     void Format() ;
                                     // Does a low-level format of the store.

                                     char* Get_Name() ;
                                     // Returns name of store.

                                     TCOM_Cache64* Get_Cache() ;
                                     // Returns current cache object.

                                     void Set_Cache( TCOM_Cache64* Value ) ;
                                     // Sets the current cache object.

                                     bool Contiguous_Store() ;
                                     /* Returns True if the store's addresses are contiguous
                                       (non-sparse). */

                                     void Set_Max_Storage( TStore_Address64 Value,
                                         TUnified_Exception* &Res ) ;
                                     /* Sets the maximum (current) size of the store, in bytes.
                                       In other words, this is the new highest valid address
                                       (plus 1).  If the store is not resizable, the call
                                       fails and an error code is returned in Res.  This should
                                       only be used to shrink the size of the store, not to
                                       extend it.  To extend, use the Extend method. */

                                     TStore_Address64 Extended_Size() ;
                                     /* Returns the maximum theoretical size that the store can
                                       be extended to.  If the store is not extendable, this
                                       returns the same value as Max_Storage. */

                                   long int Get_Bytes_Read() ;
                                       // Data bytes read
                                   long int Get_Bytes_Written() ;
                                       // Data bytes written
                                   long int Get_Reads() ;
                                       // Data read count
                                   long int Get_Writes() ;
                                       // Data write count
                                   long int Get_Error_Count() ;
                                       // Error count
                                   void Set_Bytes_Read( long int Value ) ;
                                       // Data bytes read
                                   void Set_Bytes_Written( long int Value ) ;
                                       // Data bytes written
                                   void Set_Reads( long int Value ) ;
                                       // Data read count
                                   void Set_Writes( long int Value ) ;
                                       // Data write count
                                   void Set_Error_Count( long int Value ) ;
                                       // Error count
                                   void Set_Read_Only( bool Value ) ;
                                   // Sets the store read-only value.
                                   void Set_Write_Only( bool Value ) ;
                                   // Sets the store write-only value.
} // TStandard_COM_Store64

#endif
